#!/bin/sh
echo HELM_PLUGIN_NAME=${HELM_PLUGIN_NAME}
echo HELM_PLUGIN_DIR=${HELM_PLUGIN_DIR}
echo HELM_PLUGINS=${HELM_PLUGINS}
echo HELM_REPOSITORY_CONFIG=${HELM_REPOSITORY_CONFIG}
echo HELM_REPOSITORY_CACHE=${HELM_REPOSITORY_CACHE}
echo HELM_BIN=${HELM_BIN}
